@@ -7,6 +7,7 @@ use cid::Cid;
 use filecoin_proofs_api::seal::{
     compute_comm_d, verify_aggregate_seal_commit_proofs, verify_seal as proofs_verify_seal,
 };
+use filecoin_proofs_api::update::verify_empty_sector_update_proof;
 use filecoin_proofs_api::{self as proofs, post, seal, ProverId, PublicReplicaInfo, SectorId};
 use fvm_shared::actor::builtin::Type;
 use fvm_shared::address::Protocol;
@@ -21,6 +22,7 @@ use fvm_shared::encoding::{blake2b_256, bytes_32, to_vec, RawBytes};
 use fvm_shared::error::ErrorNumber;
 use fvm_shared::piece::{zero_piece_commitment, PaddedPieceSize};
 use fvm_shared::sector::SectorInfo;
+use fvm_shared::version::NetworkVersion;
 use fvm_shared::{ActorID, FILECOIN_PRECISION};
 use lazy_static::lazy_static;
 use rayon::iter::{IndexedParallelIterator, IntoParallelRefIterator, ParallelIterator};
@@ -399,13 +401,24 @@ where
     C: CallManager,
 {
     fn total_fil_circ_supply(&self) -> Result<TokenAmount> {
-        Ok((&self.call_manager.context().fil_vested
-            + &self.get_fil_mined()?
-            + &self.get_reserve_disbursed()?
-            - &self.get_burnt_funds()?
-            - &self.power_locked()?
-            - &self.market_locked()?)
-            .max(Zero::zero()))
+        let circ_supply = if self.network_version() <= NetworkVersion::V14 {
+            // Pre-v15 the circ supply was dynamically calculated on the Filecoin mainnet,
+            // meaning it fluctuated within an epoch (as messages were executed). This forced the FVM
+            // to do much of the calculation in order to support v14.
+            (&self.call_manager.context().circ_supply
+                + &self.get_fil_mined()?
+                + &self.get_reserve_disbursed()?
+                - &self.get_burnt_funds()?
+                - &self.power_locked()?
+                - &self.market_locked()?)
+                .max(Zero::zero())
+        } else {
+            // From v15 and onwards, Filecoin mainnet was fixed to use a static circ supply per epoch.
+            // The value reported to the FVM from clients is now the static value,
+            // the FVM simply reports that value to actors.
+            self.call_manager.context().circ_supply.clone()
+        };
+        Ok(circ_supply)
     }
 }
 
@@ -668,6 +681,26 @@ where
         )
         .or_illegal_argument()
     }
+
+    fn verify_replica_update(&mut self, replica: &ReplicaUpdateInfo) -> Result<bool> {
+        self.call_manager.charge_gas(
+            self.call_manager
+                .price_list()
+                .on_verify_replica_update(replica),
+        )?;
+
+        let up: proofs::RegisteredUpdateProof =
+            replica.update_proof_type.try_into().or_illegal_argument()?;
+
+        let commr_old =
+            cid_to_replica_commitment_v1(&replica.old_sealed_cid).or_illegal_argument()?;
+        let commr_new =
+            cid_to_replica_commitment_v1(&replica.new_sealed_cid).or_illegal_argument()?;
+        let commd = cid_to_data_commitment_v1(&replica.new_unsealed_cid).or_illegal_argument()?;
+
+        verify_empty_sector_update_proof(up, &replica.proof, commr_old, commr_new, commd)
+            .or_illegal_argument()
+    }
 }
 
 impl<C> GasOps for DefaultKernel<C>
