@@ -1,3 +1,5 @@
+use std::ops::RangeInclusive;
+
 use anyhow::{anyhow, Context as _};
 use cid::Cid;
 use fvm_shared::actor::builtin::Manifest;
@@ -56,26 +58,29 @@ where
         engine: Engine,
         epoch: ChainEpoch,
         base_fee: TokenAmount,
-        fil_vested: TokenAmount,
+        circ_supply: TokenAmount,
         network_version: NetworkVersion,
         state_root: Cid,
         builtin_actors: Cid,
         blockstore: B,
         externs: E,
     ) -> anyhow::Result<Self> {
+        const SUPPORTED_VERSIONS: RangeInclusive<NetworkVersion> =
+            NetworkVersion::V14..=NetworkVersion::V15;
+
         debug!(
             "initializing a new machine, epoch={}, base_fee={}, nv={:?}, root={}",
             epoch, &base_fee, network_version, state_root
         );
 
-        if network_version != NetworkVersion::V14 {
+        if !SUPPORTED_VERSIONS.contains(&network_version) {
             return Err(anyhow!("unsupported network version: {}", network_version));
         }
 
         let context = MachineContext {
             epoch,
             base_fee,
-            fil_vested,
+            circ_supply,
             network_version,
             initial_state_root: state_root,
             price_list: price_list_by_epoch(epoch),
@@ -94,6 +99,7 @@ where
         }
 
         // Load the built-in actors manifest.
+        // TODO: Check that the actor bundle is sane for the network version.
         let builtin_actors: Manifest = blockstore
             .get_cbor(&builtin_actors)
             .context("failed to load built-in actor index")?
@@ -196,41 +202,84 @@ where
     }
 
     fn transfer(&mut self, from: ActorID, to: ActorID, value: &TokenAmount) -> Result<()> {
-        if from == to || value.is_zero() {
-            return Ok(());
-        }
-
-        if value.is_negative() {
-            return Err(syscall_error!(IllegalArgument;
+        if self.context.network_version >= NetworkVersion::V15 {
+            if value.is_negative() {
+                return Err(syscall_error!(IllegalArgument;
                 "attempted to transfer negative transfer value {}", value)
-            .into());
-        }
+                .into());
+            }
 
-        // If the from actor doesn't exist, we return "insufficient funds" to distinguish between
-        // that and the case where the _receiving_ actor doesn't exist.
-        let mut from_actor = self
-            .state_tree
-            .get_actor_id(from)?
-            .context("cannot transfer from non-existent sender")
-            .or_error(ErrorNumber::InsufficientFunds)?;
+            // If the from actor doesn't exist, we return "insufficient funds" to distinguish between
+            // that and the case where the _receiving_ actor doesn't exist.
+            let mut from_actor = self
+                .state_tree
+                .get_actor_id(from)?
+                .context("cannot transfer from non-existent sender")
+                .or_error(ErrorNumber::InsufficientFunds)?;
 
-        let mut to_actor = self
-            .state_tree
-            .get_actor_id(to)?
-            .context("cannot transfer to non-existent receiver")
-            .or_error(ErrorNumber::NotFound)?;
+            if from_actor.balance.lt(value) {
+                return Err(syscall_error!(InsufficientFunds; "sender does not have funds to transfer (balance {}, transfer {})", &from_actor.balance, value).into());
+            }
 
-        from_actor.deduct_funds(value).map_err(|e| {
-            syscall_error!(InsufficientFunds;
+            if from == to {
+                debug!("attempting to self-transfer: noop (from/to: {})", from);
+                return Ok(());
+            }
+
+            let mut to_actor = self
+                .state_tree
+                .get_actor_id(to)?
+                .context("cannot transfer to non-existent receiver")
+                .or_error(ErrorNumber::NotFound)?;
+
+            from_actor.deduct_funds(value).map_err(|e| {
+                syscall_error!(InsufficientFunds;
                            "transfer failed when deducting funds ({}) from balance ({}): {}",
                            value, &from_actor.balance, e)
-        })?;
-        to_actor.deposit_funds(value);
+            })?;
+            to_actor.deposit_funds(value);
 
-        self.state_tree.set_actor_id(from, from_actor)?;
-        self.state_tree.set_actor_id(to, to_actor)?;
+            self.state_tree.set_actor_id(from, from_actor)?;
+            self.state_tree.set_actor_id(to, to_actor)?;
 
-        log::trace!("transfered {} from {} to {}", value, from, to);
+            log::trace!("transferred {} from {} to {}", value, from, to);
+        } else {
+            if from == to || value.is_zero() {
+                return Ok(());
+            }
+
+            if value.is_negative() {
+                return Err(syscall_error!(IllegalArgument;
+                "attempted to transfer negative transfer value {}", value)
+                .into());
+            }
+
+            // If the from actor doesn't exist, we return "insufficient funds" to distinguish between
+            // that and the case where the _receiving_ actor doesn't exist.
+            let mut from_actor = self
+                .state_tree
+                .get_actor_id(from)?
+                .context("cannot transfer from non-existent sender")
+                .or_error(ErrorNumber::InsufficientFunds)?;
+
+            let mut to_actor = self
+                .state_tree
+                .get_actor_id(to)?
+                .context("cannot transfer to non-existent receiver")
+                .or_error(ErrorNumber::NotFound)?;
+
+            from_actor.deduct_funds(value).map_err(|e| {
+                syscall_error!(InsufficientFunds;
+                           "transfer failed when deducting funds ({}) from balance ({}): {}",
+                           value, &from_actor.balance, e)
+            })?;
+            to_actor.deposit_funds(value);
+
+            self.state_tree.set_actor_id(from, from_actor)?;
+            self.state_tree.set_actor_id(to, to_actor)?;
+
+            log::trace!("transfered {} from {} to {}", value, from, to);
+        }
 
         Ok(())
     }
